@@ -22,6 +22,7 @@
 """Defines interface for DB access."""
 
 import datetime
+import itertools
 import threading
 
 from oslo_config import cfg
@@ -31,9 +32,6 @@ from oslo_log import log as logging
 from oslo_utils import excutils
 import osprofiler.sqlalchemy
 from retrying import retry
-import six
-# NOTE(jokke): simplified transition to py3, behaves like py2 xrange
-from six.moves import range
 import sqlalchemy
 from sqlalchemy.ext.compiler import compiles
 from sqlalchemy import MetaData, Table
@@ -74,7 +72,7 @@ def _retry_on_deadlock(exc):
     """Decorator to retry a DB API call if Deadlock was received."""
 
     if isinstance(exc, db_exception.DBDeadlock):
-        LOG.warn(_LW("Deadlock detected. Retrying..."))
+        LOG.warning(_LW("Deadlock detected. Retrying..."))
         return True
     return False
 
@@ -131,7 +129,7 @@ def clear_db_env():
 
 def _check_mutate_authorization(context, image_ref):
     if not is_image_mutable(context, image_ref):
-        LOG.warn(_LW("Attempted to modify image user did not own."))
+        LOG.warning(_LW("Attempted to modify image user did not own."))
         msg = _("You do not own this image")
         if image_ref.visibility in ['private', 'shared']:
             exc_class = exception.Forbidden
@@ -364,7 +362,7 @@ def _paginate_query(query, model, limit, sort_keys, marker=None,
     if 'id' not in sort_keys:
         # TODO(justinsb): If this ever gives a false-positive, check
         # the actual primary key, rather than assuming its id
-        LOG.warn(_LW('Id not in sort_keys; is sort_keys unique?'))
+        LOG.warning(_LW('Id not in sort_keys; is sort_keys unique?'))
 
     assert(not (sort_dir and sort_dirs))  # nosec
     # nosec: This function runs safely if the assertion fails.
@@ -746,6 +744,74 @@ def _image_get_disk_usage_by_owner(owner, session, image_id=None):
     return total
 
 
+def _image_get_staging_usage_by_owner(owner, session):
+    # NOTE(danms): We could do this in a single query, but I think it is
+    # easier to understand as two that we concatenate while generating
+    # results.
+
+    # Images in uploading or importing state are consuming staging
+    # space.
+    query = session.query(models.Image)
+    query = query.filter(models.Image.owner == owner)
+    query = query.filter(models.Image.size > 0)
+    query = query.filter(models.Image.status.in_(('uploading',
+                                                  'importing')))
+    importing_images = query.all()
+
+    # Images with non-empty os_glance_importing_to_stores properties
+    # may also be consuming staging space. Filter our deleted images
+    # or importing ones included in the above query.
+    props = session.query(models.ImageProperty).filter(
+        models.ImageProperty.name == 'os_glance_importing_to_stores',
+        models.ImageProperty.value != '').subquery()
+    query = session.query(models.Image)
+    query = query.join(props, props.c.image_id == models.Image.id)
+    query = query.filter(models.Image.owner == owner)
+    query = query.filter(models.Image.size > 0)
+    query = query.filter(~models.Image.status.in_(('uploading',
+                                                   'importing',
+                                                   'killed',
+                                                   'deleted')))
+    copying_images = query.all()
+
+    return sum(i.size for i in itertools.chain(importing_images,
+                                               copying_images))
+
+
+def _image_get_count_by_owner(owner, session):
+    query = session.query(models.Image)
+    query = query.filter(models.Image.owner == owner)
+    query = query.filter(~models.Image.status.in_(['killed', 'deleted']))
+    return query.count()
+
+
+def _image_get_uploading_count_by_owner(owner, session):
+    """Return a count of the images uploading or importing."""
+
+    importing_statuses = ('saving', 'uploading', 'importing')
+
+    # Images in any state indicating uploading, including through image_upload
+    # or importing count for this.
+    query = session.query(models.Image)
+    query = query.filter(models.Image.owner == owner)
+    query = query.filter(models.Image.status.in_(importing_statuses))
+    uploading = query.count()
+
+    # Images that are not in the above list, but are not deleted and
+    # in the process of doing a copy count for this.
+    props = session.query(models.ImageProperty).filter(
+        models.ImageProperty.name == 'os_glance_importing_to_stores',
+        models.ImageProperty.value != '').subquery()
+    query = session.query(models.Image)
+    query = query.join(props, props.c.image_id == models.Image.id)
+    query = query.filter(models.Image.owner == owner)
+    query = query.filter(~models.Image.status.in_(importing_statuses +
+                                                  ('killed', 'deleted')))
+    copying = query.count()
+
+    return uploading + copying
+
+
 def _validate_image(values, mandatory_status=True):
     """
     Validates the incoming data and raises a Invalid exception
@@ -804,13 +870,13 @@ def image_set_property_atomic(image_id, name, value):
         table = models.ImageProperty.__table__
 
         # This should be:
-        #   UPDATE image_properties SET value=$value, deleted=0
-        #     WHERE name=$name AND deleted!=0
+        #   UPDATE image_properties SET value=$value, deleted=False
+        #     WHERE name=$name AND deleted!=False
         result = connection.execute(table.update().where(
             sa_sql.and_(table.c.name == name,
                         table.c.image_id == image_id,
-                        table.c.deleted != 0)).values(
-                            value=value, deleted=0))
+                        table.c.deleted != False)).values(
+                            value=value, deleted=False))
         if result.rowcount == 1:
             # Found and updated a deleted property, so we win
             return
@@ -855,7 +921,7 @@ def image_delete_property_atomic(image_id, name, value):
             sa_sql.and_(table.c.name == name,
                         table.c.value == value,
                         table.c.image_id == image_id,
-                        table.c.deleted == 0)))
+                        table.c.deleted == False)))
         if result.rowcount == 1:
             return
 
@@ -1018,7 +1084,7 @@ def image_location_update(context, image_id, location, session=None):
     except sa_orm.exc.NoResultFound:
         msg = (_("No location found with ID %(loc)s from image %(img)s") %
                dict(loc=loc_id, img=image_id))
-        LOG.warn(msg)
+        LOG.warning(msg)
         raise exception.NotFound(msg)
 
 
@@ -1044,7 +1110,7 @@ def image_location_delete(context, image_id, location_id, status,
     except sa_orm.exc.NoResultFound:
         msg = (_("No location found with ID %(loc)s from image %(img)s") %
                dict(loc=location_id, img=image_id))
-        LOG.warn(msg)
+        LOG.warning(msg)
         raise exception.NotFound(msg)
 
 
@@ -1106,7 +1172,7 @@ def _set_properties_for_image(context, image_ref, properties,
     for prop_ref in image_ref.properties:
         orig_properties[prop_ref.name] = prop_ref
 
-    for name, value in six.iteritems(properties):
+    for name, value in properties.items():
         prop_values = {'image_id': image_ref.id,
                        'name': name,
                        'value': value}
@@ -1182,12 +1248,20 @@ def _image_property_update(context, prop_ref, values, session=None):
 
 def image_property_delete(context, prop_ref, image_ref, session=None):
     """
-    Used internally by image_property_create and image_property_update.
+    Used internally by _set_properties_for_image().
     """
     session = session or get_session()
     prop = session.query(models.ImageProperty).filter_by(image_id=image_ref,
                                                          name=prop_ref).one()
-    prop.delete(session=session)
+    try:
+        prop.delete(session=session)
+    except sa_orm.exc.StaleDataError as e:
+        LOG.debug(('StaleDataError while deleting property %(prop)r '
+                   'from image %(image)r likely means we raced during delete: '
+                   '%(err)s'),
+                  {'prop': prop_ref, 'image': image_ref,
+                   'err': str(e)})
+        return
     return prop
 
 
@@ -1450,7 +1524,7 @@ def purge_deleted_rows(context, age_in_days, max_rows, session=None):
     except (db_exception.DBError, db_exception.DBReferenceError) as ex:
         LOG.exception(_LE('DBError detected when force purging '
                           'table=%(table)s: %(error)s'),
-                      {'table': ti, 'error': six.text_type(ex)})
+                      {'table': ti, 'error': str(ex)})
         raise
 
     rows = result.rowcount
@@ -1495,7 +1569,7 @@ def purge_deleted_rows(context, age_in_days, max_rows, session=None):
             with excutils.save_and_reraise_exception():
                 LOG.error(_LE('DBError detected when purging from '
                           "%(tablename)s: %(error)s"),
-                          {'tablename': tbl, 'error': six.text_type(ex)})
+                          {'tablename': tbl, 'error': str(ex)})
 
         rows = result.rowcount
         LOG.info(_LI('Deleted %(rows)d row(s) from table %(tbl)s'),
@@ -1539,7 +1613,7 @@ def purge_deleted_rows_from_images(context, age_in_days, max_rows,
         with excutils.save_and_reraise_exception():
             LOG.error(_LE('DBError detected when purging from '
                       "%(tablename)s: %(error)s"),
-                      {'tablename': tbl, 'error': six.text_type(ex)})
+                      {'tablename': tbl, 'error': str(ex)})
 
     rows = result.rowcount
     LOG.info(_LI('Deleted %(rows)d row(s) from table %(tbl)s'),
@@ -1554,6 +1628,21 @@ def user_get_storage_usage(context, owner_id, image_id=None, session=None):
     return total_size
 
 
+def user_get_staging_usage(context, owner_id, session=None):
+    session = session or get_session()
+    return _image_get_staging_usage_by_owner(owner_id, session)
+
+
+def user_get_image_count(context, owner_id, session=None):
+    session = session or get_session()
+    return _image_get_count_by_owner(owner_id, session)
+
+
+def user_get_uploading_count(context, owner_id, session=None):
+    session = session or get_session()
+    return _image_get_uploading_count_by_owner(owner_id, session)
+
+
 def _task_info_format(task_info_ref):
     """Format a task info ref for consumption outside of this module"""
     if task_info_ref is None:
@@ -1661,6 +1750,44 @@ def task_get(context, task_id, session=None, force_show_deleted=False):
     return _task_format(task_ref, task_ref.info)
 
 
+def tasks_get_by_image(context, image_id, session=None):
+    """Fetch all tasks associated with image_id"""
+    tasks = []
+    session = session or get_session()
+    _task_soft_delete(context, session=session)
+
+    query = session.query(models.Task).options(
+        sa_orm.joinedload(models.Task.info)
+    ).filter_by(image_id=image_id)
+
+    expires_at = models.Task.expires_at
+    query = query.filter(sa_sql.or_(expires_at == None,
+                                    expires_at >= timeutils.utcnow()))
+    updated_at = models.Task.updated_at
+    query.filter(
+        updated_at <= (timeutils.utcnow() +
+                       datetime.timedelta(hours=CONF.task.task_time_to_live)))
+
+    if not context.can_see_deleted:
+        query = query.filter_by(deleted=False)
+
+    try:
+        task_refs = query.all()
+    except sa_orm.exc.NoResultFound:
+        LOG.debug("No task found for image with ID %s", image_id)
+        return tasks
+
+    for task_ref in task_refs:
+        # Make sure the task is visible
+        if not _is_task_visible(context, task_ref):
+            msg = "Task %s is not visible, excluding" % task_ref.id
+            LOG.debug(msg)
+            continue
+        tasks.append(_task_format(task_ref, task_ref.info))
+
+    return tasks
+
+
 def task_delete(context, task_id, session=None):
     """Delete a task"""
     session = session or get_session()
@@ -1807,7 +1934,10 @@ def _task_format(task_ref, task_info_ref=None):
         'created_at': task_ref['created_at'],
         'updated_at': task_ref['updated_at'],
         'deleted_at': task_ref['deleted_at'],
-        'deleted': task_ref['deleted']
+        'deleted': task_ref['deleted'],
+        'image_id': task_ref['image_id'],
+        'request_id': task_ref['request_id'],
+        'user_id': task_ref['user_id'],
     }
 
     if task_info_ref:
@@ -2050,11 +2180,11 @@ def metadef_tag_create(context, namespace_name, tag_dict,
 
 
 def metadef_tag_create_tags(context, namespace_name, tag_list,
-                            session=None):
+                            can_append=False, session=None):
     """Create a metadata-schema tag or raise if it already exists."""
     session = get_session()
     return metadef_tag_api.create_tags(
-        context, namespace_name, tag_list, session)
+        context, namespace_name, tag_list, can_append, session)
 
 
 @utils.no_4byte_params
